#include "eventLog.h"

//////////////////////////////////////////////////////////////////////////////


using namespace std;
typedef list<SourceData> SourceDataList;


//
//  FUNCTION: int GetNewRecords( char* logName, 
//								 SourceDataList	&srcList,				   
//								 DWORD nSec )
//
//  PURPOSE: Get all logrecords added to the specified eventlog file 
//			 during the last N seconds (inclusive description strings)
//
//  PARAMETERS:
//		logName - name of the logfile to open
//		srcList - list which new eventlogrecords will be added to 
//		nSec - number of seconds to "look back" in the eventlog
//		nErr - number
//		
//  RETURN VALUES:
//		number of records added to the list if the funtion returned successfully		
//		-1 if the function did not succeed
//
int GetNewRecords( char*			logName, 
				   SourceDataList	&srcList,				   
				   DWORD			nSec );


//
//  FUNCTION: BuildEventLogBuffer( SourceDataList	&srcList, 
//								   EventLogLimits*	lim,
//								   char*			buffer,
//								   DWORD			bufSize)
//
//  PURPOSE: build a buffer of eventlog records which has exceeded any of
//			 the limits specified.  This is the buffer which is sent to the client
//
//  PARAMETERS:
//		srcList - list which contains eventlog records
//		lim	- limits which the eventlog records cannot exceed
//		buffer - buffer of records that has exceeded any of the limits given				 
//
//  RETURN VALUES:
//		none
//
//	COMMENT:
//	the buffer has the following format:
// --------------------------------------------------------------------------------------------
// |nbytes=N;cname=COMPUTER;stat=STATUS;src=SOURCENAME;id=ID;type=TYPE;cnt=COUNT;msg=MESSAGE;;|
// --------------------------------------------------------------------------------------------
//				
// nbytes - number of bytes in the buffer
// cname - name of this computer
// stat - status of the eventlog (based on the limits)
//		  could be 'error', 'warning' or 'ok'
// src - name of the source which has has broken any of the limits
// id - the event id of the record
// type - message type (information, error, warning, auditFailure or auditSuccess)
// cnt - number of messages of this message id
// msg - the description string of this record
//
// the buffer can contain multiple records if neccessary
//
void BuildEventLogBuffer( SourceDataList	&srcList, 
						  EventLogLimits*	lim,
						  char*				buffer,
						  DWORD				bufSize);

//
//  FUNCTION: GetBytesNeededForEventLogBuffer
//
//  PURPOSE: Compute the number of bytes needed to build a
//			 buffer of eventlog records
//
//  PARAMETERS:
//		srcList - list which contains eventlog records
//		lim	- limits which the eventlog records cannot exceed
//		nBytesNeeded - conatins the number of bytes needed when the
//					   function has returned
//  RETURN VALUES:
//		none		
//	
void GetBytesNeededForEventLogBuffer( const SourceDataList	&srcList,									 
									  EventLogLimits*		lim,
									  int					&nBytesNeeded);


//////////////////////////////////////////////////////////////////////////////


void AddEventSource(char *installDir)
{
    HKEY hk; 
    DWORD dwData; 
    char*	szBuf;
	int		bufSize;	
	char*	regKey = "SYSTEM\\CurrentControlSet\\Services\\EventLog\\Application\\";
 
	if(strlen(regKey) > strlen(installDir))
		bufSize =   strlen(regKey) + strlen(SZAPPNAME);	
	else
		bufSize =   strlen(installDir+1) + strlen(SZAPPNAME);	

	szBuf = new char[bufSize+4];

	strcpy(szBuf, regKey);	
	strcat(szBuf, SZAPPNAME);

    // Add your source name as a subkey under the Application 
    // key in the EventLog registry key. 
 
	if (RegCreateKey(HKEY_LOCAL_MACHINE, szBuf, &hk)) {
		PrintToFile("ERR: Could not create the registry key", ERRFILE);
		delete szBuf;
		return;
	}     
 
    // Set the name of the message file

    strcpy(szBuf, installDir);
	strcat(szBuf, "\\");
	strcat(szBuf, SZAPPNAME);
	strcat(szBuf, ".exe");
 
    // Add the name to the EventMessageFile subkey. 
 
    if (RegSetValueEx(hk,             // subkey handle 
            "EventMessageFile",       // value name 
            0,                        // must be zero 
            REG_EXPAND_SZ,            // value type 
            (LPBYTE) szBuf,      // pointer to value data 
            strlen(szBuf) + 1))  // length of value data 
	{
		PrintToFile("ERR: Could not set the event message file", ERRFILE);
		RegCloseKey(hk);
		delete szBuf;
		return;	
	}       
 
    // Set the supported event types in the TypesSupported subkey. 
 
    dwData = EVENTLOG_ERROR_TYPE | EVENTLOG_WARNING_TYPE | 
		     EVENTLOG_INFORMATION_TYPE; 
 
    if (RegSetValueEx(hk,      // subkey handle 
            "TypesSupported",  // value name 
            0,                 // must be zero 
            REG_DWORD,         // value type 
            (LPBYTE) &dwData,  // pointer to value data 
            sizeof(DWORD)))    // length of value data 
	{
		PrintToFile("ERR: Could not set the supported types", ERRFILE);		
	}	

	
	delete szBuf;	
    RegCloseKey(hk);
} 


void AddToMessageLog(LPSTR szMsg, WORD wType)
{
    HANDLE h;
	const char*  Strings[1];

	Strings[0] = szMsg;
 
    h = RegisterEventSource( NULL,			// uses local computer 
							 SZAPPNAME );   // source name 
    if (h == NULL) {
		PrintToFile("ERR: Could not register the event source", ERRFILE);
        return;
	}
 
    if ( !ReportEvent( h,				  // event log handle 
					   wType,			  // event type 
					   0,                 // category zero 
					   MSG_GENERAL,		  // event identifier 
					   NULL,              // no user security identifier 
					   1,                 // one substitution strings 
					   0,                 // no data 
					   Strings,			  // pointer to string array 
					   NULL) ) {          // pointer to data 
        
		PrintToFile("ERR: Could not report the event", ERRFILE);
        return;
	}
 
    DeregisterEventSource(h); 
} 


bool GetDescriptionString(char				*sourceName,
						  char				*logName, 
						  EVENTLOGRECORD	*record, 
						  char				*descStr, 
						  int				&bufSize)
{
	int			nBytes=0;
	int			totBytes=0;
	bool		ret=true;
	int			res;
	HKEY		hRegKey = NULL;	
	char		*pSubKey=NULL;
	int			strLen;
	char	    *szBuf=NULL;	// buffer to contain messagefile-name
	DWORD		szBufSize;	
	char		*str1 = "SYSTEM\\CurrentControlSet\\Services\\EventLog\\";	
	HMODULE		hModule = NULL;	
	LPTSTR		lpszTemp = NULL;
	DWORD		dwFormatFlags;
	char		**msgTable=NULL;
	int			offset;
	int			i;
		
	strLen = strlen(sourceName) + strlen(logName) + strlen(str1);

	pSubKey = new char[strLen+10];

	strcpy(pSubKey, str1);
	strcat(pSubKey, logName);
	strcat(pSubKey, "\\");
	strcat(pSubKey, sourceName);
	
	res = RegOpenKey( HKEY_LOCAL_MACHINE,	// handle to open key
					  pSubKey,				// address of name of subkey to open
					  &hRegKey );			// address of handle to open key

	
	if( res != ERROR_SUCCESS) 
	{
		ret = false;
		goto cleanup;
	}

	// get required bufferspace
	res =  RegQueryValueEx( hRegKey,			 // handle to key to query
							"EventMessageFile",  // address of name of value to query
							0,					 // reserved
							NULL,				 // address of buffer for value type
							NULL,			  	 // address of data buffer
							&szBufSize);		 // address of data buffer size
	
	if( res != ERROR_SUCCESS) 
	{
		ret = false;
		goto cleanup;
	}
	
	szBuf = new char[szBufSize];

	// get name of messagefile
	res =  RegQueryValueEx( hRegKey,			 // handle to key to query
							"EventMessageFile",  // address of name of value to query
							0,					 // reserved
							NULL,				 // address of buffer for value type
							(LPBYTE) szBuf,      // address of data buffer
							&szBufSize);		 // address of data buffer size

	if( res != ERROR_SUCCESS) 
	{
		ret = false;
		goto cleanup;
	}

	if( szBuf[0]=='%' ) // "%SystemRoot%"
	{
		char buf[256];		
		char *tmp;
		GetWindowsDirectory(buf, 256);

		tmp = new char[strlen(buf)+strlen(szBuf)];

		strcpy(tmp, buf);
		strcat(tmp, &szBuf[12]);
		
		delete szBuf;
		szBuf = new char[strlen(tmp)+1];
		strcpy(szBuf, tmp);
	
		delete tmp;
	}
	
	// open the messagefile	
	hModule = LoadLibraryEx(szBuf, NULL, LOAD_LIBRARY_AS_DATAFILE);	

	DWORD error;
	error = GetLastError();
							  

	if(hModule==NULL) 
	{		
		ret = false;
		msgTable = new char*[1];
		msgTable[0] = " ";
		bufSize = 1;
		goto cleanup;
	}
	
	dwFormatFlags = FORMAT_MESSAGE_ALLOCATE_BUFFER |						  						  
	 			    FORMAT_MESSAGE_FROM_HMODULE |
					FORMAT_MESSAGE_ARGUMENT_ARRAY;


	msgTable = new char*[record->NumStrings];
	offset=0;
	
	// insert the insertion strings from the eventrecord into msgTable
	for(i=0; i<record->NumStrings; i++)	
	{
		nBytes = strlen((char*) ((LPBYTE) record + record->StringOffset + offset));
		nBytes++;		
		totBytes+=nBytes;		
		msgTable[i] = new char[nBytes];		
		strncpy(msgTable[i], (char*) ((LPBYTE) record + record->StringOffset + offset), nBytes);		
		offset += strlen(msgTable[i])+1;
	}
	
	// get the description string
	res = FormatMessage( dwFormatFlags,		 // source and processing options
						 hModule,			 // pointer to  message source
						 record->EventID,	 // requested message identifier
						 MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // language identifier for requested message
						 (LPTSTR)&lpszTemp,	 // pointer to message buffer
						 0,					 // maximum size of message buffer
						 msgTable );
	
	if ( res == NULL || bufSize < strlen(lpszTemp) )
	{
		bufSize = strlen(lpszTemp);
		ret = false;
		goto cleanup;
	}		
	
	if(lpszTemp[strlen(lpszTemp)-2]==13)	// ascii 13 = carriage return
		lpszTemp[strlen(lpszTemp)-2]='\0';

	strcpy(descStr, lpszTemp);	

cleanup:	
	if(lpszTemp!=NULL)
  		LocalFree((HLOCAL) lpszTemp );	
	if(hModule!=NULL)
		FreeLibrary(hModule);	
	if(hRegKey!=NULL)
		CloseHandle(hRegKey);	
	if(pSubKey!=NULL)
		delete pSubKey;
	if(szBuf!=NULL)
		delete szBuf;
	if(msgTable!=NULL && nBytes>0)
	{
		for(i=0;i<record->NumStrings;i++)			    
			delete msgTable[i];		
		delete msgTable;
	}

	return ret;
}

int GetNewRecords( char*			logName, 
				   SourceDataList	&srcList,				   
				   DWORD			nSec )
				   
{	
	SourceDataList::iterator it;

    HANDLE			h;
    EVENTLOGRECORD	*pevlr; 
    BYTE			bBuffer[1000]; 
    DWORD			dwRead, dwNeeded, dwThisRecord = 0; 
	bool			firstTime = true;
	bool			stop = false;
	bool			listUpdated=false;	
	int				nNewRecords = 0;  
	int				bytesRequired=0;
	bool			res;	
	SourceData		srcData;    
	srcData.message = NULL;	
	int				nListElements=0;
	DWORD			dwTime;
	
	h = OpenEventLog( NULL,       // use local computer 
					  logName);   // source name 
    if (h == NULL) 
        return -1;	

    pevlr = (EVENTLOGRECORD *) &bBuffer;    
	
	time( (time_t*) &dwTime );
	dwTime -= nSec;	
	
    while (ReadEventLog( h,							// event log handle 
						 EVENTLOG_BACKWARDS_READ |	// reads forward 
						 EVENTLOG_SEQUENTIAL_READ,	// sequential read 
						 0,							// ignored for sequential reads 
						 pevlr,						// pointer to buffer 
						 1000,						// size of buffer 
						 &dwRead,					// number of bytes read 
						 &dwNeeded)					// bytes in next record 
			&& !stop )
	{
        while (dwRead > 0) 
        {
			if( dwTime <= pevlr->TimeGenerated)
			{				
				
				strncpy(srcData.sourcName, (LPSTR) ((LPBYTE) pevlr + sizeof(EVENTLOGRECORD)), 255);
				srcData.EventType = pevlr->EventType;
				srcData.eventID = pevlr->EventID;				
			
				// run through the list and check if the combination
				// sourceName - eventId already exist
				for (it = srcList.begin(); it != srcList.end(); ++it)
				{
					if( strcmp(it->sourcName, srcData.sourcName) == 0
						&& it->eventID == srcData.eventID )
					{					
						switch(srcData.EventType)
						{
						case EVENTLOG_ERROR_TYPE:							
							it->nErr++;							
							break;
						case EVENTLOG_WARNING_TYPE:			
							it->nWarn++;
							break;
						case EVENTLOG_INFORMATION_TYPE:				
							it->nInf++;					
							break;
						case EVENTLOG_AUDIT_SUCCESS:			
							it->nAudS++;
							break;
						case EVENTLOG_AUDIT_FAILURE:			
							it->nAudF++;
							break;						
						}
					
						it->nMessages++;							
						listUpdated = true;											
					}
				}				

				// A new message with a new id has been found
				if(!listUpdated) 
				{

					srcData.nErr = srcData.nWarn = srcData.nInf = 0;
					srcData.nAudF = srcData.nAudS = 0;								

					switch(srcData.EventType)
					{
					case EVENTLOG_ERROR_TYPE:
						srcData.nErr++;
						break;
					case EVENTLOG_WARNING_TYPE:			
						srcData.nWarn++;
						break;
					case EVENTLOG_INFORMATION_TYPE:				
						srcData.nInf++;
						break;
					case EVENTLOG_AUDIT_SUCCESS:			
						srcData.nAudS++;
						break;
					case EVENTLOG_AUDIT_FAILURE:			
						srcData.nAudF++;
						break;			
					}
										
					srcData.nMessages = 1;									
					srcList.insert (srcList.begin(), srcData);
					it = srcList.begin();
									
					bytesRequired = 0;		

					// Get number of bytes required for the buffer
					GetDescriptionString((char*) ((LPBYTE) pevlr + sizeof(EVENTLOGRECORD)), logName, pevlr, it->message, bytesRequired);
					it->message = new char[bytesRequired];				

					// Get the description string
					res = GetDescriptionString((char*) ((LPBYTE) pevlr + sizeof(EVENTLOGRECORD)), logName, pevlr, it->message, bytesRequired);
					nListElements++;			  	  
					
				}	

				listUpdated = false;				
				nNewRecords++;
			}
			else 
			{
				stop=true;
				break;
			}			
            
			dwRead -= pevlr->Length; 
            pevlr = (EVENTLOGRECORD *) ((LPBYTE) pevlr + pevlr->Length); 
        } 
 
        pevlr = (EVENTLOGRECORD *) &bBuffer; 
    }	

    CloseEventLog(h);
	return nNewRecords;
}


void BuildEventLogBuffer( SourceDataList	&srcList, 
						  EventLogLimits*	lim,
						  char*				buffer,
						  DWORD				bufSize)
{
	SourceDataList::iterator it;

	int			nBytesNeeded=0;
	char		status;	
	DWORD		computerNameSize=256;	
	char		computerName[256];
	char		tmp[20];
	bool		err=false;

	bool		ok=true;
	char		stat[15];	
	
	// bytes needed
	strcpy(buffer, "nbytes=");
	strcat(buffer, _itoa(bufSize, tmp, 10));

	// get gomputer Name	
	GetComputerName( computerName,		// address of name buffer
					 &computerNameSize  // address of size of name buffer
					);
	
	strcat(buffer, ";cname=");
	strcat(buffer, computerName);
	strcat(buffer, ";");

	// check the status
	strcpy(stat, "stat=ok;");
	
	for (it = srcList.end(); it != srcList.begin(); --it) 
	{
		status = CheckEventLogStat(it->EventType, it->nMessages, lim);	

		if(status == 'e')
		{
			strcpy(stat, "stat=error;");
			break;
		}
		else if(status == 'w')
			strcpy(stat, "stat=warning;");
	}	
	
	if( strncmp(stat, "stat=ok", 7) == 0 ) 
	{
		strcpy(buffer, stat);
		return;
	}

	strcat(buffer, stat);

	// find application(s) which has broken any of the limits
	for (it = srcList.begin(); it != srcList.end(); ++it) 
	{		
		if( CheckEventLogStat(it->EventType, it->nMessages, lim) != 'o')						
		{						
			// sourcename
			strcat(buffer, "src=");
			strcat(buffer, it->sourcName);
			strcat(buffer, ";");
			// eventID
			strcat(buffer, "id=");			
			strcat(buffer, _ultoa( (DWORD) (LOWORD(it->eventID)), tmp, 10));
		    strcat(buffer, ";");			
			// eventType
			strcat(buffer, "type=");
			strcat(buffer, _ultoa( it->EventType, tmp, 10));
			strcat(buffer, ";");
			// message count
			strcat(buffer, "cnt=");
			strcat(buffer, _ultoa( it->nMessages, tmp, 10));
			strcat(buffer, ";");			
			// message			
			strcat(buffer, "msg=");
			strcat(buffer, it->message);
			strcat(buffer, ";");		
		}
	}	
	
	// end
	strcat(buffer, ";");	
}


void GetBytesNeededForEventLogBuffer( SourceDataList	&srcList,									 
									  EventLogLimits*	lim,
									  int				&nBytesNeeded)
{
	char	*computerName=0;
	DWORD	bufSize=0;
	char	tmp[10];
	SourceDataList::iterator it;
	bool err=false;

	
	nBytesNeeded = strlen("stat=warning;");

	// get buffersize for Computer Name
	GetComputerName( computerName,  // address of name buffer
					 &bufSize       // address of size of name buffer
					);

	// "cname=" + computername
	nBytesNeeded += (bufSize + 6);

	for (it = srcList.end(); it != srcList.begin(); --it) 
	{
		if( CheckEventLogStat(it->EventType, it->nMessages, lim) != 'o')						
		{			
			// start ";src=": 5 byte
			nBytesNeeded += 5;		
			// sourcename + ";"
			nBytesNeeded += strlen(it->sourcName) + 1;
			// "id=" + eventID + ";"
			nBytesNeeded += strlen( _ultoa( (DWORD) (LOBYTE(LOWORD(it->eventID))), tmp, 10)) + 4;
			// "type=" + eventType + ";"
			nBytesNeeded += 6;
			// "cnt=" + message count + ";"
			nBytesNeeded += strlen( _ultoa( it->nMessages, tmp, 10)) + 5;			
			// "msg=" + message + ";"
			nBytesNeeded += strlen(it->message) + 5;
			err=true;
		}		
	}
	
	if(!err)
	{
		nBytesNeeded = 8;
		return;
	}

	// stop ";;": 2 byte
	nBytesNeeded += 2;

	// "nbytes=" + N + ";"
	nBytesNeeded += strlen("nbytes=") + strlen(_itoa(nBytesNeeded, tmp, 10)) + 10;
}


bool CheckEventLog( char	*ibuf,
					char	*buffer,
					DWORD	&bufSize)
{
	int				nNewRecords=0;
	int				applog;
	int				seclog;
	int				syslog;		
	int				nBytesNeeded=0;	
	int				period;	
	int				index=0;
	bool			ret=true;
	EventLogLimits	limits;
	
	SourceDataList	srcDataList;
	SourceDataList::iterator it;	

	period = GetEventlogData(ibuf, "per=");
	applog = GetEventlogData(ibuf, "app=");
	seclog = GetEventlogData(ibuf, "sec=");
	syslog = GetEventlogData(ibuf, "sys=");

	limits.wAudF = GetEventlogData(ibuf, "wAudF=");
	limits.eAudF = GetEventlogData(ibuf, "eAudF=");
	limits.wAudS = GetEventlogData(ibuf, "wAudS=");
	limits.eAudS = GetEventlogData(ibuf, "eAudS=");
	limits.wErr  = GetEventlogData(ibuf, "wErr=");
	limits.eErr  = GetEventlogData(ibuf, "eErr=");
	limits.wWarn = GetEventlogData(ibuf, "wWarn=");
	limits.eWarn = GetEventlogData(ibuf, "eWarn=");
	limits.wInf  = GetEventlogData(ibuf, "wInf=");
	limits.eInf  = GetEventlogData(ibuf, "eInf=");
		
	// get number of elements needed for scrDataArr	
	if(applog) 
	{		
		 GetNewRecords( "Application", 
						srcDataList,						 
						period*60 );
	}
	
	if(seclog)
	{		
		GetNewRecords( "Security", 
						srcDataList,						
						period*60 );		
	}	
	
	if(syslog) 
	{		
		GetNewRecords( "System", 
						srcDataList,						
						period*60 );		
	}
	
	GetBytesNeededForEventLogBuffer( srcDataList, &limits, nBytesNeeded );
	
	if(bufSize < (DWORD)nBytesNeeded) 
	{
		bufSize = (DWORD)nBytesNeeded;
		ret = false;
	}
	else
	{	
		cout << "\nBytes Needed " << nBytesNeeded << endl;	

		BuildEventLogBuffer( srcDataList, &limits, buffer, nBytesNeeded );	

		cout << "Bytes used: " << strlen(buffer) << endl;	
	}

	for (it = srcDataList.begin(); it != srcDataList.end(); ++it) 
		delete it->message;

	srcDataList.clear();
	
	return ret;
}

char CheckEventLogStat(DWORD EventType, int nErr, EventLogLimits *lim) 
{
	switch(EventType)
	{
	case EVENTLOG_ERROR_TYPE:
		if(lim->wErr == 0 || nErr < lim->wErr )
			return 'o';
		if(nErr > lim->eErr)
			return 'e';		
		return 'w';		
	case EVENTLOG_WARNING_TYPE:
		if(lim->wWarn == 0 || nErr < lim->wWarn )
			return 'o';
		if(nErr > lim->eWarn)
			return 'e';		
		return 'w';		
	case EVENTLOG_INFORMATION_TYPE:
		if(lim->wInf == 0 || nErr < lim->wInf )
			return 'o';
		if(nErr > lim->eInf)
			return 'e';			
		return 'w';		
	case EVENTLOG_AUDIT_SUCCESS:
		if(lim->wAudS == 0 || nErr < lim->wAudS )
			return 'o';
		if(nErr > lim->eAudS)
			return 'e';		
		return 'w';		
	case EVENTLOG_AUDIT_FAILURE:
		if(lim->wAudF == 0 || nErr < lim->wAudF )
			return 'o';
		if(nErr > lim->eAudF)
			return 'e';		
		return 'w';	
	}

	return 'o';
}

int GetEventlogData(char *ibuf, char *label)
{	

	DWORD	index = 0;
	int		lim=0;

	while( strncmp(&ibuf[index], label, strlen(label))!=0 && index < strlen(ibuf))
		index++;

	index+=strlen(label);

	return atoi(&ibuf[index]);	
}